from typing import List, Union

from openai import AsyncOpenAI, OpenAI

from greptimeai import logger
from greptimeai.collector import Collector
from greptimeai.patcher import Patcher
from greptimeai.patcher.openai_patcher.base import (
    _AudioPatcher,
    _ChatCompletionPatcher,
    _CompletionPatcher,
    _EmbeddingPatcher,
    _FilePatcher,
    _FineTuningPatcher,
    _ImagePatcher,
    _ModelPatcher,
    _ModerationPatcher,
)
from greptimeai.patcher.openai_patcher.retry import _RetryPatcher

_collector: Collector = None  # type: ignore


def setup(
    host: str = "",
    database: str = "",
    token: str = "",
    client: Union[OpenAI, AsyncOpenAI, None] = None,
):
    """
    patch openai functions automatically.

    host, database and token is to setup the place to store the data, and the authority.
    They MUST BE set explicitly by passing parameters or system environment variables.

    Args:
        host: if None or empty string, GREPTIMEAI_HOST environment variable will be used.
        database: if None or empty string, GREPTIMEAI_DATABASE environment variable will be used.
        token: if None or empty string, GREPTIMEAI_TOKEN environment variable will be used.
        client: if None, then openai module-level client will be patched.
    """
    global _collector
    _collector = Collector(
        service_name="openai", host=host, database=database, token=token
    )
    patchers: List[Patcher] = [
        _AudioPatcher(collector=_collector, client=client),
        _ChatCompletionPatcher(collector=_collector, client=client),
        _CompletionPatcher(collector=_collector, client=client),
        _EmbeddingPatcher(collector=_collector, client=client),
        _FilePatcher(collector=_collector, client=client),
        _FineTuningPatcher(collector=_collector, client=client),
        _ImagePatcher(collector=_collector, client=client),
        _ModelPatcher(collector=_collector, client=client),
        _ModerationPatcher(collector=_collector, client=client),
        _RetryPatcher(collector=_collector, client=client),
    ]

    for patcher in patchers:
        patcher.patch()

    logger.info("ready to track openai metrics and traces.")



from typing import Any, Dict, Optional, Tuple

from opentelemetry.util.types import Attributes

from greptimeai.labels import (
    _COMPLETION_COST_LABEL,
    _COMPLETION_TOKENS_LABEL,
    _ERROR_TYPE_LABEL,
    _MODEL_LABEL,
    _SPAN_NAME_LABEL,
    _PROMPT_COST_LABEl,
    _PROMPT_TOKENS_LABEl,
)

from .collection import _Collector


class Collector:
    """
    tracker to collect metrics and traces
    """

    def __init__(
        self, service_name: str, host: str = "", database: str = "", token: str = ""
    ):
        self._collector = _Collector(
            service_name=service_name, host=host, database=database, token=token
        )

    def start_span(
        self,
        span_name: str,
        span_attrs: Dict[str, Any],
        event_name: str,
        event_attrs: Dict[str, Any],
    ) -> Tuple[str, str]:
        return self._collector.start_span(
            span_id=None,
            parent_id=None,
            span_name=span_name,
            event_name=event_name,
            span_attrs=span_attrs,
            event_attrs=event_attrs,
        )

    def end_span(
        self,
        span_id: str,
        span_name: str,
        span_attrs: Dict[str, Any],
        event_attrs: Dict[str, Any],
        ex: Optional[Exception] = None,
    ):
        self._collector.end_span(
            span_id=span_id,
            span_name=span_name,
            event_name="end",
            span_attrs=span_attrs,
            event_attrs=event_attrs,
            ex=ex,
        )

    def collect_error_count(
        self, model_name: Optional[str], span_name: str, ex: Exception
    ):
        attributes = {
            _ERROR_TYPE_LABEL: ex.__class__.__name__,
            _SPAN_NAME_LABEL: span_name,
        }
        if model_name:
            attributes[_MODEL_LABEL] = model_name

        self._collector.collect_error_count(attributes=attributes)

    def collect_metrics(self, span_attrs: Dict[str, Any], attrs: Optional[Attributes]):
        """
        Collects metrics for the given span attributes and optional attributes.

        Args:
            span_attrs (Dict[str, Any]): The span attributes containing the metrics data.
            attrs (Optional[Attributes]): Optional additional attributes.

        Notes:
            The `attrs` for metrics should be small enough to keep the series small.
            Detailed information should be stored in traces instead of metrics.
        """
        prompt_tokens = span_attrs.get(_PROMPT_TOKENS_LABEl, 0)
        prompt_cost = span_attrs.get(_PROMPT_COST_LABEl, 0)
        completion_tokens = span_attrs.get(_COMPLETION_TOKENS_LABEL, 0)
        completion_cost = span_attrs.get(_COMPLETION_COST_LABEL, 0)

        self._collector.collect_metrics(
            prompt_tokens=prompt_tokens,
            prompt_cost=prompt_cost,
            completion_tokens=completion_tokens,
            completion_cost=completion_cost,
            attrs=attrs,
        )



from typing import Any, Dict, Iterable, Optional, Sequence, Union

from langchain.schema import ChatGeneration, Generation
from langchain.schema.document import Document
from langchain.schema.messages import BaseMessage

_SPAN_NAME_CHAIN = "langchain_chain"
_SPAN_NAME_AGENT = "langchain_agent"
_SPAN_NAME_LLM = "langchain_llm"
_SPAN_NAME_TOOL = "langchain_tool"
_SPAN_NAME_RETRIEVER = "langchain_retriever"


def _get_user_id(metadata: Optional[Dict[str, Any]]) -> str:
    """
    get user id from metadata
    """
    return (metadata or {}).get("user_id", "")


def _get_serialized_id(serialized: Dict[str, Any]) -> Optional[str]:
    """
    get id if exist
    """
    ids = serialized.get("id")
    if ids and isinstance(ids, list):
        return ids[len(ids) - 1]
    return None


def _get_serialized_streaming(serialized: Dict[str, Any]) -> bool:
    """
    get streaming if exist
    """
    id = _get_serialized_id(serialized)
    if not id:
        return False

    if id == "OpenAI" or id == "ChatOpenAI":
        return serialized.get("kwargs", {}).get("streaming")
    return False


def _parse(obj: Any) -> Union[Dict[str, Any], Sequence[Any], Any]:
    if hasattr(obj, "to_json"):
        return obj.to_json()

    if isinstance(obj, dict):
        return {key: _parse(value) for key, value in obj.items()}

    if isinstance(obj, list):
        return [_parse(item) for item in obj]

    return obj


def _parse_input(raw_input: Any) -> Any:
    if not raw_input:
        return None

    if not isinstance(raw_input, dict):
        return _parse(raw_input)

    return (
        raw_input.get("input")
        or raw_input.get("inputs")
        or raw_input.get("question")
        or raw_input.get("query")
        or _parse(raw_input)
    )


def _parse_output(raw_output: dict) -> Any:
    if not raw_output:
        return None

    if not isinstance(raw_output, dict):
        return _parse(raw_output)

    return (
        raw_output.get("text")
        or raw_output.get("output")
        or raw_output.get("output_text")
        or raw_output.get("answer")
        or raw_output.get("result")
        or _parse(raw_output)
    )


def _parse_generation(gen: Generation) -> Optional[Dict[str, Any]]:
    """
    Generation, or ChatGeneration (which contains message field)
    """
    if not gen:
        return None

    info = gen.generation_info or {}
    attrs = {
        "text": gen.text,
        # the following is OpenAI only?
        "finish_reason": info.get("finish_reason"),
        "log_probability": info.get("logprobs"),
    }

    if isinstance(gen, ChatGeneration):
        message: BaseMessage = gen.message
        attrs["additional_kwargs"] = message.additional_kwargs
        attrs["type"] = message.type

    return attrs


def _parse_generations(
    gens: Sequence[Generation],
) -> Optional[Iterable[Dict[str, Any]]]:
    """
    parse LLMResult.generations[0] to structured fields
    """
    if gens and len(gens) > 0:
        return list(filter(None, [_parse_generation(gen) for gen in gens if gen]))

    return None


def _parse_documents(docs: Sequence[Document]) -> Optional[Sequence[Dict[str, Any]]]:
    """
    parse LLMResult.generations[0] to structured fields
    """

    def _parse_doc(doc: Document) -> Dict[str, Any]:
        return {
            "content": doc.page_content,
            "metadata": doc.metadata,
        }

    if docs and len(docs) > 0:
        return [_parse_doc(doc) for doc in docs]

    return None



import time
from typing import Any, AsyncIterator, Dict, Iterator, Optional, Tuple, Union

from openai import AsyncStream, Stream
from openai.types.chat import ChatCompletionChunk
from openai.types.completion import Completion

from greptimeai import logger
from greptimeai.collector import Collector
from greptimeai.labels import (
    _COMPLETION_COST_LABEL,
    _COMPLETION_TOKENS_LABEL,
    _MODEL_LABEL,
    _OUTPUT_DISPLAY_LABEL,
    _SPAN_NAME_LABEL,
)
from greptimeai.utils.openai.token import (
    get_openai_token_cost_for_model,
    num_tokens_from_messages,
)


def _extract_resp(resp: Any) -> Dict[str, Any]:
    try:
        if hasattr(resp, "model_dump"):
            return resp.model_dump()
        else:
            logger.warning(f"Unknown response stream type: {type(resp)}")
            return {}
    except Exception as e:
        logger.error(f"Failed to extract response: {e}")
        return {}


def _extract_chat_completion_chunk_tokens(chunk: ChatCompletionChunk) -> str:
    if not chunk:
        return ""

    tokens = ""
    try:
        for choice in chunk.choices:
            if choice.delta.content:
                tokens += choice.delta.content
    except Exception as e:
        logger.error(f"Failed to extract chat completion chunk tokens: {e}")

    return tokens


def _extract_completion_tokens(completion: Completion) -> str:
    if not completion:
        return ""

    tokens = ""

    try:
        for choice in completion.choices:
            tokens += choice.text
    except Exception as e:
        logger.error(f"Failed to extract completion tokens: {e}")

    return tokens


def _extract_tokens(resp: Any) -> str:
    if isinstance(resp, ChatCompletionChunk):
        return _extract_chat_completion_chunk_tokens(resp)
    elif isinstance(resp, Completion):
        return _extract_completion_tokens(resp)
    else:
        logger.warning(f"Unsupported response stream type: {type(resp)}")
        return ""


def _collect_resp(
    resp: Any, collector: Collector, span_id: str, event_name: str
) -> Tuple[str, str]:
    event_attrs = _extract_resp(resp)
    model_name = event_attrs.get("model", "")
    tokens = _extract_tokens(resp)
    collector._collector.add_span_event(
        span_id=span_id,
        event_name=event_name,
        event_attrs=event_attrs,
    )

    return tokens, model_name


def _end_collect(
    collector: Collector,
    span_id: str,
    span_name: str,
    model_name: str,
    start: float,
    tokens: str,
):
    num = num_tokens_from_messages(tokens)
    cost = get_openai_token_cost_for_model(model_name, num)

    span_attrs = {
        _COMPLETION_TOKENS_LABEL: num,
        _COMPLETION_COST_LABEL: cost,
        _MODEL_LABEL: model_name,
        _OUTPUT_DISPLAY_LABEL: tokens,
    }

    attrs: Dict[str, Union[str, bool]] = {
        _SPAN_NAME_LABEL: f"{span_name}[stream]",
        _MODEL_LABEL: model_name,
    }

    latency = 1000 * (time.time() - start)

    collector.collect_metrics(span_attrs=span_attrs, attrs=attrs)

    collector._collector.record_latency(latency, attributes=attrs)

    collector.end_span(
        span_id=span_id,
        span_name=span_name,
        span_attrs=span_attrs,
        event_attrs={},
    )


class Stream_(Stream):
    def __init__(
        self,
        stream: Stream,
        collector: Collector,
        span_id: str,
        span_name: str,
        start: float,
        model_name: Optional[str],
    ):
        self.stream = stream
        self.collector = collector
        self.span_id = span_id
        self.span_name = span_name
        self.start = start
        self.model_name = model_name or ""

    def __iter__(self) -> Iterator[Any]:
        completion_tokens = ""

        for item in self.stream:
            yield item

            tokens, model_name = _collect_resp(
                item, self.collector, self.span_id, "stream"
            )
            completion_tokens += tokens
            if model_name:
                self.model_name = model_name

        _end_collect(
            collector=self.collector,
            span_id=self.span_id,
            span_name=self.span_name,
            model_name=self.model_name,
            start=self.start,
            tokens=completion_tokens,
        )


class AsyncStream_(AsyncStream):
    def __init__(
        self,
        astream: AsyncStream,
        collector: Collector,
        span_id: str,
        span_name: str,
        start: float,
        model_name: Optional[str],
    ):
        self.astream = astream
        self.collector = collector
        self.span_id = span_id
        self.span_name = span_name
        self.start = start
        self.model_name = model_name or ""

    async def __aiter__(self) -> AsyncIterator[Any]:
        completion_tokens = ""

        async for item in self.astream:
            yield item

            tokens, model_name = _collect_resp(
                item, self.collector, self.span_id, "stream"
            )
            completion_tokens += tokens
            if model_name:
                self.model_name = model_name

        _end_collect(
            collector=self.collector,
            span_id=self.span_id,
            span_name=self.span_name,
            model_name=self.model_name,
            start=self.start,
            tokens=completion_tokens,
        )



from typing import Union

import openai
from openai import AsyncOpenAI, OpenAI

from greptimeai.patchee import Patchee

from . import _SPAN_NAME_MODEL, OpenaiPatchees


class _ModelPatchees:
    def __init__(self, client: Union[OpenAI, AsyncOpenAI, None], method_name: str):
        models = Patchee(
            obj=client.models if client else openai.models,
            method_name=method_name,
            span_name=_SPAN_NAME_MODEL,
            event_name=f"models.{method_name}",
        )

        models_raw = Patchee(
            obj=client.models.with_raw_response
            if client
            else openai.models.with_raw_response,
            method_name=method_name,
            span_name=_SPAN_NAME_MODEL,
            event_name=f"models.with_raw_response.{method_name}",
        )

        self.patchees = [models, models_raw]

        if client:
            raw_models = Patchee(
                obj=client.with_raw_response.models,
                method_name=method_name,
                span_name=_SPAN_NAME_MODEL,
                event_name=f"with_raw_response.models.{method_name}",
            )
            self.patchees.append(raw_models)


class _ModelListPatchees(_ModelPatchees):
    def __init__(self, client: Union[OpenAI, AsyncOpenAI, None] = None):
        super().__init__(client=client, method_name="list")


class _ModelRetrievePatchees(_ModelPatchees):
    def __init__(self, client: Union[OpenAI, AsyncOpenAI, None] = None):
        super().__init__(client=client, method_name="retrieve")


class _ModelDeletePatchees(_ModelPatchees):
    def __init__(self, client: Union[OpenAI, AsyncOpenAI, None] = None):
        super().__init__(client=client, method_name="delete")


class ModelPatchees(OpenaiPatchees):
    def __init__(self, client: Union[OpenAI, AsyncOpenAI, None] = None):
        list = _ModelListPatchees(client=client)
        retrieve = _ModelRetrievePatchees(client=client)
        delete = _ModelDeletePatchees(client=client)

        patchees = list.patchees + retrieve.patchees + delete.patchees

        super().__init__(patchees=patchees, client=client)



from typing import Union

import openai
from openai import AsyncOpenAI, OpenAI

from greptimeai.patchee import Patchee

from . import _SPAN_NAME_IMAGE, OpenaiPatchees


class _ImagePatchees:
    def __init__(self, client: Union[OpenAI, AsyncOpenAI, None], method_name: str):
        images = Patchee(
            obj=client.images if client else openai.images,
            method_name=method_name,
            span_name=_SPAN_NAME_IMAGE,
            event_name=f"images.{method_name}",
        )

        images_raw = Patchee(
            obj=client.images.with_raw_response
            if client
            else openai.images.with_raw_response,
            method_name=method_name,
            span_name=_SPAN_NAME_IMAGE,
            event_name=f"images.with_raw_response.{method_name}",
        )

        self.patchees = [images, images_raw]

        if client:
            raw_images = Patchee(
                obj=client.with_raw_response.images,
                method_name=method_name,
                span_name=_SPAN_NAME_IMAGE,
                event_name=f"with_raw_response.images.{method_name}",
            )
            self.patchees.append(raw_images)


class _ImageGeneratePatchees(_ImagePatchees):
    def __init__(self, client: Union[OpenAI, AsyncOpenAI, None] = None):
        super().__init__(client=client, method_name="generate")


class _ImageEditPatchees(_ImagePatchees):
    def __init__(self, client: Union[OpenAI, AsyncOpenAI, None] = None):
        super().__init__(client=client, method_name="edit")


class _ImageVariationPatchees(_ImagePatchees):
    def __init__(self, client: Union[OpenAI, AsyncOpenAI, None] = None):
        super().__init__(client=client, method_name="create_variation")


class ImagePatchees(OpenaiPatchees):
    def __init__(self, client: Union[OpenAI, AsyncOpenAI, None] = None):
        generate = _ImageGeneratePatchees(client=client)
        edit = _ImageEditPatchees(client=client)
        variation = _ImageVariationPatchees(client=client)

        patchees = generate.patchees + edit.patchees + variation.patchees

        super().__init__(patchees=patchees, client=client)



import functools
import time
from typing import Any, Callable, Dict, Optional, Tuple, Union

from openai import AsyncOpenAI, AsyncStream, OpenAI, Stream
from typing_extensions import override

from greptimeai import logger
from greptimeai.collector import Collector
from greptimeai.extractor import Extraction
from greptimeai.extractor.openai_extractor import OpenaiExtractor
from greptimeai.labels import (
    _MODEL_LABEL,
    _SPAN_NAME_LABEL,
    _PROMPT_COST_LABEl,
    _PROMPT_TOKENS_LABEl,
)
from greptimeai.patchee import Patchee
from greptimeai.patchee.openai_patchee import OpenaiPatchees
from greptimeai.patchee.openai_patchee.audio import AudioPatchees
from greptimeai.patchee.openai_patchee.chat_completion import ChatCompletionPatchees
from greptimeai.patchee.openai_patchee.completion import CompletionPatchees
from greptimeai.patchee.openai_patchee.embedding import EmbeddingPatchees
from greptimeai.patchee.openai_patchee.file import FilePatchees
from greptimeai.patchee.openai_patchee.fine_tuning import FineTuningPatchees
from greptimeai.patchee.openai_patchee.image import ImagePatchees
from greptimeai.patchee.openai_patchee.model import ModelPatchees
from greptimeai.patchee.openai_patchee.moderation import ModerationPatchees
from greptimeai.patcher import Patcher
from greptimeai.patcher.openai_patcher.stream import AsyncStream_, Stream_
from greptimeai.utils.openai.token import (
    get_openai_token_cost_for_model,
    num_tokens_from_messages,
)


class _OpenaiPatcher(Patcher):
    def __init__(
        self,
        patchees: OpenaiPatchees,  # specify what methods to be patched
        collector: Collector,  # collect metrics and traces
        extractor: Optional[OpenaiExtractor] = None,  # extract info from req and resp
        client: Union[OpenAI, AsyncOpenAI, None] = None,
    ):
        self.collector = collector
        self.extractor = extractor or OpenaiExtractor()
        self._is_async = False
        if isinstance(client, AsyncOpenAI):
            self._is_async = True

        prefix = "client" if client else "openai"
        suffix = "[async]" if self._is_async else ""
        for patchee in patchees.get_patchees():
            patchee.event_name = f"{prefix}.{patchee.event_name}{suffix}"

        self.patchees = patchees

    def _collect_req_metrics_for_stream(
        self, model_name: str, span_name: str, tokens_num: int, cost: float
    ):
        attrs = {
            _SPAN_NAME_LABEL: f"{span_name}[stream]",
            _MODEL_LABEL: model_name,
        }

        span_attrs = {
            _PROMPT_TOKENS_LABEl: tokens_num,
            _PROMPT_COST_LABEl: cost,
        }

        self.collector.collect_metrics(span_attrs=span_attrs, attrs=attrs)

    def _pre_patch(
        self,
        patchee: Patchee,
        *args,
        **kwargs,
    ) -> Tuple[Extraction, str, float, Dict[str, Any]]:
        extraction = self.extractor.pre_extract(*args, **kwargs)

        # if stream, the usage won't be included in the resp,
        # so we need to extract and collect it from req for best.
        if OpenaiExtractor.is_stream(**kwargs):
            tokens = OpenaiExtractor.extract_req_tokens(**kwargs)
            model_name = extraction.get_model_name() or ""
            tokens_num = num_tokens_from_messages(tokens or "")
            cost = get_openai_token_cost_for_model(model_name, tokens_num, False)

            extraction = OpenaiExtractor.supplement_stream_prompt(
                extraction, tokens_num, cost
            )

            self._collect_req_metrics_for_stream(
                model_name=model_name,
                span_name=patchee.span_name,
                tokens_num=tokens_num,
                cost=cost,
            )

        trace_id, span_id = self.collector.start_span(
            span_name=patchee.span_name,
            event_name=patchee.event_name,
            span_attrs=extraction.span_attributes,
            event_attrs=extraction.event_attributes,
        )
        OpenaiExtractor.update_trace_info(kwargs, trace_id, span_id)

        start = time.time()
        return (extraction, span_id, start, kwargs)

    def _post_patch(
        self,
        span_id: str,
        start: float,
        span_name: str,
        resp: Any,
        ex: Optional[Exception] = None,
    ):
        latency = 1000 * (time.time() - start)
        extraction = self.extractor.post_extract(resp)
        attrs = {_SPAN_NAME_LABEL: span_name}
        model = extraction.get_model_name()
        if model:
            attrs[_MODEL_LABEL] = model

        self.collector._collector.record_latency(latency, attributes=attrs)
        self.collector.end_span(
            span_id=span_id,
            span_name=span_name,
            span_attrs=extraction.span_attributes,
            event_attrs=extraction.event_attributes,
            ex=ex,
        )
        self.collector.collect_metrics(
            span_attrs=extraction.span_attributes, attrs=attrs
        )

    def _patch_sync(self, func: Callable, patchee: Patchee):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            extraction, span_id, start, kwargs = self._pre_patch(
                patchee, *args, **kwargs
            )
            resp, ex = None, None
            try:
                resp = func(*args, **kwargs)
            except Exception as e:
                self.collector.collect_error_count(
                    extraction.get_model_name(), patchee.span_name, e
                )
                ex = e
                raise e
            finally:
                if isinstance(resp, Stream):
                    resp = Stream_(
                        stream=resp,
                        collector=self.collector,
                        span_id=span_id,
                        span_name=patchee.span_name,
                        start=start,
                        model_name=extraction.get_model_name(),
                    )
                else:
                    self._post_patch(
                        span_id=span_id,
                        start=start,
                        span_name=patchee.span_name,
                        resp=resp,
                        ex=ex,
                    )
            return resp

        patchee.wrap_func(wrapper)
        logger.debug(f"patched '{patchee}'")

    def _patch_async(self, func: Callable, patchee: Patchee):
        """
        NOTE: this is exactly the same as _patch_sync, but with async/await
        """

        @functools.wraps(func)
        async def async_wrapper(*args, **kwargs):
            extraction, span_id, start, kwargs = self._pre_patch(
                patchee, *args, **kwargs
            )
            resp, ex = None, None
            try:
                resp = await func(*args, **kwargs)
            except Exception as e:
                self.collector.collect_error_count(
                    extraction.get_model_name(), patchee.span_name, e
                )
                ex = e
                raise e
            finally:
                if isinstance(resp, AsyncStream):
                    resp = AsyncStream_(
                        astream=resp,
                        collector=self.collector,
                        span_id=span_id,
                        span_name=patchee.span_name,
                        start=start,
                        model_name=extraction.get_model_name(),
                    )
                else:
                    self._post_patch(
                        span_id=span_id,
                        start=start,
                        span_name=patchee.span_name,
                        resp=resp,
                        ex=ex,
                    )
            return resp

        patchee.wrap_func(async_wrapper)
        logger.debug(f"patched '{patchee}'")

    @override
    def patch(self):
        for patchee in self.patchees.get_patchees():
            func = patchee.get_unwrapped_func()
            if not func:
                return

            if self._is_async:
                self._patch_async(func, patchee)
            else:
                self._patch_sync(func, patchee)


class _AudioPatcher(_OpenaiPatcher):
    def __init__(
        self,
        collector: Collector,
        client: Union[OpenAI, AsyncOpenAI, None] = None,
    ):
        patchees = AudioPatchees(client=client)
        super().__init__(collector=collector, patchees=patchees, client=client)


class _ChatCompletionPatcher(_OpenaiPatcher):
    def __init__(
        self,
        collector: Collector,
        client: Union[OpenAI, AsyncOpenAI, None] = None,
    ):
        patchees = ChatCompletionPatchees(client=client)
        super().__init__(collector=collector, patchees=patchees, client=client)


class _CompletionPatcher(_OpenaiPatcher):
    def __init__(
        self,
        collector: Collector,
        client: Union[OpenAI, AsyncOpenAI, None] = None,
    ):
        patchees = CompletionPatchees(client=client)
        super().__init__(collector=collector, patchees=patchees, client=client)


class _EmbeddingPatcher(_OpenaiPatcher):
    def __init__(
        self,
        collector: Collector,
        client: Union[OpenAI, AsyncOpenAI, None] = None,
    ):
        patchees = EmbeddingPatchees(client=client)
        super().__init__(collector=collector, patchees=patchees, client=client)


class _FilePatcher(_OpenaiPatcher):
    def __init__(
        self,
        collector: Collector,
        client: Union[OpenAI, AsyncOpenAI, None] = None,
    ):
        patchees = FilePatchees(client=client)
        super().__init__(collector=collector, patchees=patchees, client=client)


class _FineTuningPatcher(_OpenaiPatcher):
    def __init__(
        self,
        collector: Collector,
        client: Union[OpenAI, AsyncOpenAI, None] = None,
    ):
        patchees = FineTuningPatchees(client=client)
        super().__init__(collector=collector, patchees=patchees, client=client)


class _ImagePatcher(_OpenaiPatcher):
    def __init__(
        self,
        collector: Collector,
        client: Union[OpenAI, AsyncOpenAI, None] = None,
    ):
        patchees = ImagePatchees(client=client)
        super().__init__(collector=collector, patchees=patchees, client=client)


class _ModelPatcher(_OpenaiPatcher):
    def __init__(
        self,
        collector: Collector,
        client: Union[OpenAI, AsyncOpenAI, None] = None,
    ):
        patchees = ModelPatchees(client=client)
        super().__init__(collector=collector, patchees=patchees, client=client)


class _ModerationPatcher(_OpenaiPatcher):
    def __init__(
        self,
        collector: Collector,
        client: Union[OpenAI, AsyncOpenAI, None] = None,
    ):
        patchees = ModerationPatchees(client=client)
        super().__init__(collector=collector, patchees=patchees, client=client)



from typing import Any, Dict, Optional, Tuple, Union

from openai._response import APIResponse
from pydantic import BaseModel
from typing_extensions import override

from greptimeai import logger
from greptimeai.extractor import BaseExtractor, Extraction
from greptimeai.labels import (
    _COMPLETION_COST_LABEL,
    _COMPLETION_TOKENS_LABEL,
    _INPUT_DISPLAY_LABEL,
    _MODEL_LABEL,
    _OUTPUT_DISPLAY_LABEL,
    _USER_ID_LABEL,
    _PROMPT_COST_LABEl,
    _PROMPT_TOKENS_LABEl,
)
from greptimeai.utils.openai.token import (
    extract_chat_inputs,
    extract_chat_outputs,
    get_openai_token_cost_for_model,
)

_EXTRA_HEADERS_KEY = "extra_headers"  # this is the original openai parameter
_EXTRA_HEADERS_X_USER_ID_KEY = "x-user-id"
_EXTRA_HEADERS_X_TRACE_ID_KEY = "x-trace-id"
_EXTRA_HEADERS_X_SPAN_ID_KEY = "x-span-id"

_GREPTIMEAI_USER_KEY = "user_id"
_GREPTIMEAI_TRACE_KEY = "trace_id"
_GREPTIMEAI_SPAN_KEY = "span_id"

_OPENAI_USER_KEY = "user"  # this is the original openai parameter


class OpenaiExtractor(BaseExtractor):
    @staticmethod
    def is_stream(**kwargs) -> bool:
        stream = kwargs.get("stream")
        if isinstance(stream, bool):
            return stream
        elif isinstance(stream, str):
            return stream.lower() == "true"
        else:
            return False

    @staticmethod
    def extract_req_tokens(**kwargs) -> Optional[Union[str, list]]:
        """
        NOTE: only for completion and chat completion so far.
        TODO(ynanbohan): better way to extract req tokens
        """
        if kwargs.get("prompt"):
            prompt = kwargs["prompt"]
            if isinstance(prompt, str):
                return prompt
            elif isinstance(prompt, list) and all(isinstance(p, str) for p in prompt):
                return " ".join(prompt)
            else:
                logger.warning(f"Failed to extract req tokens from {prompt=}")
                return None
        elif kwargs.get("messages"):
            return kwargs["messages"]
        else:
            logger.warning(f"Failed to extract req tokens from {kwargs=}")
            return None

    @staticmethod
    def update_trace_info(kwargs: Dict[str, Any], trace_id: str, span_id: str):
        attrs = {
            _EXTRA_HEADERS_X_TRACE_ID_KEY: trace_id,
            _EXTRA_HEADERS_X_SPAN_ID_KEY: span_id,
        }

        extra_headers: Dict[str, Any] = kwargs.get(_EXTRA_HEADERS_KEY, {})
        extra_headers.update(attrs)

        kwargs[_EXTRA_HEADERS_KEY] = extra_headers

    @staticmethod
    def get_trace_info(**kwargs) -> Optional[Tuple[str, str]]:
        """
        NOTE: key in headers has higher priority than the key in kwargs
        """
        trace_id = kwargs.get(_GREPTIMEAI_TRACE_KEY, "")
        span_id = kwargs.get(_GREPTIMEAI_SPAN_KEY, "")

        extra_headers = kwargs.get(_EXTRA_HEADERS_KEY)
        if extra_headers:
            trace_id = extra_headers.get(_EXTRA_HEADERS_X_TRACE_ID_KEY, "")
            span_id = extra_headers.get(_EXTRA_HEADERS_X_SPAN_ID_KEY, "")

        if trace_id or span_id:
            return trace_id, span_id

        return None

    @staticmethod
    def parse_raw_response(resp: APIResponse) -> Dict[str, Any]:
        def _http_info() -> Dict[str, Any]:
            headers = {k: v for k, v in resp.headers.items()}

            return {
                "headers": headers,
                "status_code": resp.status_code,
                "url": str(resp.url),
                "method": resp.method,
            }

        result: Dict[str, Any] = {}
        try:
            obj: BaseModel = resp.parse()
            result = obj.model_dump(exclude_unset=True)
            result["http_info"] = _http_info()
        except Exception as e:
            logger.error(f"Failed to parse response, {e}")

        return result

    @staticmethod
    def get_user_id(**kwargs) -> Optional[str]:
        """
        the upper has higher priority:
        - _X_USER_ID_KEY in extra_headers
        - _GREPTIMEAI_USER_KEY in kwargs
        - _OPENAI_USER_KEY in kwargs
        """
        user_id = kwargs.get(_GREPTIMEAI_USER_KEY) or kwargs.get(_OPENAI_USER_KEY)
        extra_headers = kwargs.get(_EXTRA_HEADERS_KEY)
        if extra_headers and _EXTRA_HEADERS_X_USER_ID_KEY in extra_headers:
            user_id = extra_headers[_EXTRA_HEADERS_X_USER_ID_KEY]
        return user_id

    @staticmethod
    def extract_usage(
        model: Optional[str], usage: Optional[Dict[str, int]]
    ) -> Dict[str, Union[float, int]]:
        res: Dict[str, Union[float, int]] = {}

        if not usage or not model:
            return res

        prompt_tokens: int = usage.get("prompt_tokens", 0)
        if prompt_tokens > 0:
            res[_PROMPT_TOKENS_LABEl] = prompt_tokens
            res[_PROMPT_COST_LABEl] = get_openai_token_cost_for_model(
                model, prompt_tokens, False
            )

        completion_tokens = usage.get("completion_tokens", 0)
        if completion_tokens > 0:
            res[_COMPLETION_TOKENS_LABEL] = completion_tokens
            res[_COMPLETION_COST_LABEL] = get_openai_token_cost_for_model(
                model, completion_tokens, True
            )
        return res

    @override
    def pre_extract(self, *args, **kwargs) -> Extraction:
        """
        extract _MODEL_LABEL, _USER_ID_LABEL for span attributes
        merge kwargs and args into event attributes
        Args:

            kwargs:
                extra_headers in kwargs: _X_USER_ID is a custom header that is used to identify the user,
                which has higher priority than the 'user' in the kwargs
        """
        span_attrs = {}
        if "model" in kwargs:
            span_attrs[_MODEL_LABEL] = kwargs["model"]

        user_id = OpenaiExtractor.get_user_id(**kwargs)
        if user_id:
            span_attrs[_USER_ID_LABEL] = user_id

        if "messages" in kwargs:
            span_attrs[_INPUT_DISPLAY_LABEL] = extract_chat_inputs(kwargs["messages"])

        event_attrs = {**kwargs}
        if len(args) > 0:
            event_attrs["args"] = args

        return Extraction(span_attributes=span_attrs, event_attributes=event_attrs)

    @override
    def post_extract(self, resp: Any) -> Extraction:
        """
        extract for span attributes:
          - _MODEL_LABEL
          - _COMPLETION_COST_LABEL
          - _COMPLETION_TOKENS_LABEL
          - _PROMPT_COST_LABEl
          - _PROMPT_TOKENS_LABEl
          - _OUTPUT_DISPLAY_LABEL

        merge usage into resp as event attributes

        Args:

            resp: inherit from the BaseModel class, or instance of APIResponse class
        """
        try:
            data: Dict[str, Any] = {}
            if isinstance(resp, APIResponse):
                data = OpenaiExtractor.parse_raw_response(resp)
                logger.debug(f"after parse_raw_response: {data=}")
            else:
                data = resp.model_dump(exclude_unset=True)
        except Exception as e:
            logger.error(f"Failed to extract response {resp}: {e}")
            data = {}

        span_attrs = {}

        model = data.get("model")
        if model:
            span_attrs[_MODEL_LABEL] = model

            usage = data.get("usage", {})
            if usage:
                usage = OpenaiExtractor.extract_usage(model, usage)
                span_attrs.update(usage)
                data["usage"] = usage

        outputs = extract_chat_outputs(data)
        if outputs:
            span_attrs[_OUTPUT_DISPLAY_LABEL] = outputs

        return Extraction(span_attributes=span_attrs, event_attributes=data)

    @staticmethod
    def supplement_stream_prompt(
        extraction: Extraction, tokens_num: int, cost: float
    ) -> Extraction:
        extraction.span_attributes[_PROMPT_TOKENS_LABEl] = tokens_num
        extraction.span_attributes[_PROMPT_COST_LABEl] = cost
        extraction.event_attributes[_PROMPT_TOKENS_LABEl] = tokens_num
        extraction.event_attributes[_PROMPT_COST_LABEl] = cost
        return extraction



from typing import Sequence, Union

from openai import AsyncOpenAI, OpenAI

from greptimeai.patchee import Patchee

_SPAN_NAME_SPEECH = "openai_speech"
_SPAN_NAME_TRANSLATION = "openai_translation"
_SPAN_NAME_TRANSCRIPTION = "openai_transcription"
_SPAN_NAME_COMPLETION = "openai_completion"
_SPAN_NAME_EMBEDDING = "openai_embedding"
_SPAN_NAME_FILE = "openai_file"
_SPAN_NAME_FINE_TUNNING = "openai_fine_tunning"
_SPAN_NAME_IMAGE = "openai_image"
_SPAN_NAME_MODEL = "openai_model"
_SPAN_NAME_MODERATION = "openai_moderation"


class OpenaiPatchees:
    patchees: Sequence[Patchee] = []
    client: Union[OpenAI, AsyncOpenAI, None]

    def __init__(
        self,
        patchees: Sequence[Patchee],
        client: Union[OpenAI, AsyncOpenAI, None] = None,
    ):
        self.patchees = patchees
        self.client = client

    def get_patchees(self) -> Sequence[Patchee]:
        return self.patchees



import functools
from typing import Callable, Optional, Union

from openai import AsyncOpenAI, OpenAI
from openai._models import FinalRequestOptions
from typing_extensions import override

from greptimeai import logger
from greptimeai.collector import Collector
from greptimeai.extractor.openai_extractor import _EXTRA_HEADERS_X_SPAN_ID_KEY
from greptimeai.patchee.openai_patchee.retry import RetryPatchees

from .base import _OpenaiPatcher


class _RetryPatcher(_OpenaiPatcher):
    def __init__(
        self,
        collector: Collector,
        client: Union[OpenAI, AsyncOpenAI, None] = None,
    ):
        self._is_async = False
        if isinstance(client, AsyncOpenAI):
            self._is_async = True
        patchees = RetryPatchees(client=client)
        super().__init__(collector=collector, patchees=patchees, client=client)

    def _add_retry_event(self, *args):
        if len(args) > 0 and isinstance(args[0], FinalRequestOptions):
            dict = args[0].model_dump(exclude_unset=True)
            span_id = dict.get("headers", {}).get(_EXTRA_HEADERS_X_SPAN_ID_KEY)
            if span_id:
                logger.debug(f"in retry_patcher {span_id=}")
                self.collector._collector.add_span_event(
                    span_id=span_id,
                    event_name="retry",
                    event_attrs=dict,
                )

    @override
    def patch(self):
        for patchee in self.patchees.get_patchees():
            func: Optional[Callable] = patchee.get_unwrapped_func()
            if func is None:
                return

            if self._is_async:

                @functools.wraps(func)
                async def async_wrapper(*args, **kwargs):
                    try:
                        self._add_retry_event(*args)
                        resp = await func(*args, **kwargs)
                    except Exception as e:
                        raise e
                    return resp

                patchee.wrap_func(async_wrapper)
                logger.debug("patched 'retry[async]'")
            else:

                @functools.wraps(func)
                def wrapper(*args, **kwargs):
                    try:
                        self._add_retry_event(*args)
                        resp = func(*args, **kwargs)
                    except Exception as e:
                        raise e
                    return resp

                patchee.wrap_func(wrapper)
                logger.debug("patched 'retry'")



from typing import Union

import openai
from openai import AsyncOpenAI, OpenAI

from greptimeai.patchee import Patchee

from . import (
    _SPAN_NAME_SPEECH,
    _SPAN_NAME_TRANSCRIPTION,
    _SPAN_NAME_TRANSLATION,
    OpenaiPatchees,
)


class _SpeechPatchees:
    def __init__(self, client: Union[OpenAI, AsyncOpenAI, None] = None):
        audio_speech_create = Patchee(
            obj=client.audio.speech if client else openai.audio.speech,
            method_name="create",
            span_name=_SPAN_NAME_SPEECH,
            event_name="audio.speech.create",
        )

        audio_raw_speech_create = Patchee(
            obj=client.audio.with_raw_response.speech
            if client
            else openai.audio.with_raw_response.speech,
            method_name="create",
            span_name=_SPAN_NAME_SPEECH,
            event_name="audio.with_raw_response.speech.create",
        )

        audio_speech_raw_create = Patchee(
            obj=client.audio.speech.with_raw_response
            if client
            else openai.audio.speech.with_raw_response,
            method_name="create",
            span_name=_SPAN_NAME_SPEECH,
            event_name="audio.speech.with_raw_response.create",
        )

        self.patchees = [
            audio_speech_create,
            audio_raw_speech_create,
            audio_speech_raw_create,
        ]

        if client:
            raw_audio_speech_create = Patchee(
                obj=client.with_raw_response.audio.speech,
                method_name="create",
                span_name=_SPAN_NAME_SPEECH,
                event_name="with_raw_response.audio.speech.create",
            )
            self.patchees.append(raw_audio_speech_create)


class _TranscriptionPatchees:
    def __init__(self, client: Union[OpenAI, AsyncOpenAI, None] = None):
        audio_transcriptions_create = Patchee(
            obj=client.audio.transcriptions if client else openai.audio.transcriptions,
            method_name="create",
            span_name=_SPAN_NAME_TRANSCRIPTION,
            event_name="audio.transcriptions.create",
        )

        audio_raw_transcriptions_create = Patchee(
            obj=client.audio.with_raw_response.transcriptions
            if client
            else openai.audio.with_raw_response.transcriptions,
            method_name="create",
            span_name=_SPAN_NAME_TRANSCRIPTION,
            event_name="audio.with_raw_response.transcriptions.create",
        )

        audio_transcriptions_raw_create = Patchee(
            obj=client.audio.transcriptions.with_raw_response
            if client
            else openai.audio.transcriptions.with_raw_response,
            method_name="create",
            span_name=_SPAN_NAME_TRANSCRIPTION,
            event_name="audio.transcriptions.with_raw_response.create",
        )

        self.patchees = [
            audio_transcriptions_create,
            audio_raw_transcriptions_create,
            audio_transcriptions_raw_create,
        ]

        if client:
            raw_audio_transcriptions_create = Patchee(
                obj=client.with_raw_response.audio.transcriptions,
                method_name="create",
                span_name=_SPAN_NAME_TRANSCRIPTION,
                event_name="with_raw_response.audio.transcriptions.create",
            )
            self.patchees.append(raw_audio_transcriptions_create)


class _TranslationPatchees:
    def __init__(self, client: Union[OpenAI, AsyncOpenAI, None] = None):
        audio_translations_create = Patchee(
            obj=client.audio.translations if client else openai.audio.translations,
            method_name="create",
            span_name=_SPAN_NAME_TRANSLATION,
            event_name="audio.translations.create",
        )

        audio_raw_translations_create = Patchee(
            obj=client.audio.with_raw_response.translations
            if client
            else openai.audio.with_raw_response.translations,
            method_name="create",
            span_name=_SPAN_NAME_TRANSLATION,
            event_name="audio.with_raw_response.translations.create",
        )

        audio_translations_raw_create = Patchee(
            obj=client.audio.translations.with_raw_response
            if client
            else openai.audio.translations.with_raw_response,
            method_name="create",
            span_name=_SPAN_NAME_TRANSLATION,
            event_name="audio.translations.with_raw_response.create",
        )

        self.patchees = [
            audio_translations_create,
            audio_raw_translations_create,
            audio_translations_raw_create,
        ]

        if client:
            raw_audio_translations_create = Patchee(
                obj=client.with_raw_response.audio.translations,
                method_name="create",
                span_name=_SPAN_NAME_TRANSLATION,
                event_name="with_raw_response.audio.translations.create",
            )
            self.patchees.append(raw_audio_translations_create)


class AudioPatchees(OpenaiPatchees):
    def __init__(self, client: Union[OpenAI, AsyncOpenAI, None] = None):
        speech = _SpeechPatchees(client)
        transcription = _TranscriptionPatchees(client)
        translation = _TranslationPatchees(client)

        patchees = speech.patchees + transcription.patchees + translation.patchees

        super().__init__(patchees=patchees, client=client)



from typing import Union

import openai
from openai import AsyncOpenAI, OpenAI

from greptimeai.patchee import Patchee

from . import _SPAN_NAME_FINE_TUNNING, OpenaiPatchees


class _FineTuningPatchees:
    def __init__(self, client: Union[OpenAI, AsyncOpenAI, None], method_name: str):
        fine_tuning_jobs = Patchee(
            obj=client.fine_tuning.jobs if client else openai.fine_tuning.jobs,
            method_name=method_name,
            span_name=_SPAN_NAME_FINE_TUNNING,
            event_name=f"fine_tuning.jobs.{method_name}",
        )

        fine_tuning_raw_jobs = Patchee(
            obj=client.fine_tuning.with_raw_response.jobs
            if client
            else openai.fine_tuning.with_raw_response.jobs,
            method_name=method_name,
            span_name=_SPAN_NAME_FINE_TUNNING,
            event_name=f"fine_tuning.with_raw_response.jobs.{method_name}",
        )

        fine_tuning_jobs_raw = Patchee(
            obj=client.fine_tuning.jobs.with_raw_response
            if client
            else openai.fine_tuning.jobs.with_raw_response,
            method_name=method_name,
            span_name=_SPAN_NAME_FINE_TUNNING,
            event_name=f"fine_tuning.jobs.with_raw_response.{method_name}",
        )

        self.patchees = [
            fine_tuning_jobs,
            fine_tuning_raw_jobs,
            fine_tuning_jobs_raw,
        ]

        if client:
            raw_fine_tuning = Patchee(
                obj=client.with_raw_response.fine_tuning.jobs,
                method_name=method_name,
                span_name=_SPAN_NAME_FINE_TUNNING,
                event_name=f"with_raw_response.fine_tuning.jobs.{method_name}",
            )
            self.patchees.append(raw_fine_tuning)


class _FineTuningListPatchees(_FineTuningPatchees):
    def __init__(self, client: Union[OpenAI, AsyncOpenAI, None] = None):
        super().__init__(client=client, method_name="list")


class _FineTuningCreatePatchees(_FineTuningPatchees):
    def __init__(self, client: Union[OpenAI, AsyncOpenAI, None] = None):
        super().__init__(client=client, method_name="create")


class _FineTuningCancelPatchees(_FineTuningPatchees):
    def __init__(self, client: Union[OpenAI, AsyncOpenAI, None] = None):
        super().__init__(client=client, method_name="cancel")


class _FineTuningRetrievePatchees(_FineTuningPatchees):
    def __init__(self, client: Union[OpenAI, AsyncOpenAI, None] = None):
        super().__init__(client=client, method_name="retrieve")


class _FineTuningListEventsPatchees(_FineTuningPatchees):
    def __init__(self, client: Union[OpenAI, AsyncOpenAI, None] = None):
        super().__init__(client=client, method_name="list_events")


class FineTuningPatchees(OpenaiPatchees):
    def __init__(self, client: Union[OpenAI, AsyncOpenAI, None] = None):
        list = _FineTuningListPatchees(client=client)
        create = _FineTuningCreatePatchees(client=client)
        cancel = _FineTuningCancelPatchees(client=client)
        retrieve = _FineTuningRetrievePatchees(client=client)
        list_events = _FineTuningListEventsPatchees(client=client)

        patchees = (
            list.patchees
            + create.patchees
            + cancel.patchees
            + retrieve.patchees
            + list_events.patchees
        )

        super().__init__(patchees=patchees, client=client)



_ERROR_TYPE_LABEL = "type"
_CLASS_TYPE_LABEL = "type"

_SPAN_NAME_LABEL = "span_name"
_USER_ID_LABEL = "user_id"
_MODEL_LABEL = "model"

_PROMPT_TOKENS_LABEl = "prompt_tokens"
_PROMPT_COST_LABEl = "prompt_cost"
_COMPLETION_TOKENS_LABEL = "completion_tokens"
_COMPLETION_COST_LABEL = "completion_cost"

# the following labels are for traces, not for metrics
_INPUT_DISPLAY_LABEL = "inputs"  # for DISPLAY, DO NOT USE IT FOR ANYTHING ELSE
_OUTPUT_DISPLAY_LABEL = "outputs"  # for DISPLAY, DO NOT USE IT FOR ANYTHING ELSE



from abc import ABC, abstractmethod
from typing import Any, Dict, Optional

from greptimeai.labels import _MODEL_LABEL


class Extraction:
    def __init__(
        self,
        span_attributes: Dict[str, Any] = {},
        event_attributes: Dict[str, Any] = {},
    ):
        self.span_attributes = span_attributes
        self.event_attributes = event_attributes

    def update_span_attributes(self, attrs: Dict[str, Any]):
        self.span_attributes.update(attrs)

    def update_event_attributes(self, attrs: Dict[str, Any]):
        self.event_attributes.update(attrs)

    def get_model_name(self) -> Optional[str]:
        return self.span_attributes.get(
            _MODEL_LABEL, None
        ) or self.event_attributes.get(_MODEL_LABEL, None)


class BaseExtractor(ABC):
    @abstractmethod
    def pre_extract(self, *args, **kwargs) -> Extraction:
        pass

    @abstractmethod
    def post_extract(self, resp: Any) -> Extraction:
        pass



from typing import Any, List, Union

from openai.types.chat import ChatCompletionMessageParam

from greptimeai import logger

# from https://github.com/langchain-ai/langchain/blob/master/libs/langchain/langchain/callbacks/openai_info.py
# 2023-11-22
MODEL_COST_PER_1K_TOKENS = {
    # GPT-4 input
    "gpt-4": 0.03,
    "gpt-4-0314": 0.03,
    "gpt-4-0613": 0.03,
    "gpt-4-32k": 0.06,
    "gpt-4-32k-0314": 0.06,
    "gpt-4-32k-0613": 0.06,
    "gpt-4-vision-preview": 0.01,
    "gpt-4-1106-preview": 0.01,
    # GPT-4 output
    "gpt-4-completion": 0.06,
    "gpt-4-0314-completion": 0.06,
    "gpt-4-0613-completion": 0.06,
    "gpt-4-32k-completion": 0.12,
    "gpt-4-32k-0314-completion": 0.12,
    "gpt-4-32k-0613-completion": 0.12,
    "gpt-4-vision-preview-completion": 0.03,
    "gpt-4-1106-preview-completion": 0.03,
    # GPT-3.5 input
    "gpt-3.5-turbo": 0.0015,
    "gpt-3.5-turbo-0301": 0.0015,
    "gpt-3.5-turbo-0613": 0.0015,
    "gpt-3.5-turbo-1106": 0.001,
    "gpt-3.5-turbo-instruct": 0.0015,
    "gpt-3.5-turbo-16k": 0.003,
    "gpt-3.5-turbo-16k-0613": 0.003,
    # GPT-3.5 output
    "gpt-3.5-turbo-completion": 0.002,
    "gpt-3.5-turbo-0301-completion": 0.002,
    "gpt-3.5-turbo-0613-completion": 0.002,
    "gpt-3.5-turbo-1106-completion": 0.002,
    "gpt-3.5-turbo-instruct-completion": 0.002,
    "gpt-3.5-turbo-16k-completion": 0.004,
    "gpt-3.5-turbo-16k-0613-completion": 0.004,
    # Azure GPT-35 input
    "gpt-35-turbo": 0.0015,  # Azure OpenAI version of ChatGPT
    "gpt-35-turbo-0301": 0.0015,  # Azure OpenAI version of ChatGPT
    "gpt-35-turbo-0613": 0.0015,
    "gpt-35-turbo-instruct": 0.0015,
    "gpt-35-turbo-16k": 0.003,
    "gpt-35-turbo-16k-0613": 0.003,
    # Azure GPT-35 output
    "gpt-35-turbo-completion": 0.002,  # Azure OpenAI version of ChatGPT
    "gpt-35-turbo-0301-completion": 0.002,  # Azure OpenAI version of ChatGPT
    "gpt-35-turbo-0613-completion": 0.002,
    "gpt-35-turbo-instruct-completion": 0.002,
    "gpt-35-turbo-16k-completion": 0.004,
    "gpt-35-turbo-16k-0613-completion": 0.004,
    # Others
    "text-ada-001": 0.0004,
    "ada": 0.0004,
    "text-babbage-001": 0.0005,
    "babbage": 0.0005,
    "text-curie-001": 0.002,
    "curie": 0.002,
    "text-davinci-003": 0.02,
    "text-davinci-002": 0.02,
    "code-davinci-002": 0.02,
    # Fine Tuned input
    "babbage-002-finetuned": 0.0016,
    "davinci-002-finetuned": 0.012,
    "gpt-3.5-turbo-0613-finetuned": 0.012,
    # Fine Tuned output
    "babbage-002-finetuned-completion": 0.0016,
    "davinci-002-finetuned-completion": 0.012,
    "gpt-3.5-turbo-0613-finetuned-completion": 0.016,
    # Azure Fine Tuned input
    "babbage-002-azure-finetuned": 0.0004,
    "davinci-002-azure-finetuned": 0.002,
    "gpt-35-turbo-0613-azure-finetuned": 0.0015,
    # Azure Fine Tuned output
    "babbage-002-azure-finetuned-completion": 0.0004,
    "davinci-002-azure-finetuned-completion": 0.002,
    "gpt-35-turbo-0613-azure-finetuned-completion": 0.002,
    # Legacy fine-tuned models
    "ada-finetuned-legacy": 0.0016,
    "babbage-finetuned-legacy": 0.0024,
    "curie-finetuned-legacy": 0.012,
    "davinci-finetuned-legacy": 0.12,
    # embedding model
    # refer: https://invertedstone.com/calculators/embedding-pricing-calculator/
    # leave embedding model *-001 cost to 0,
    "text-embedding-ada-002": 0.0004,
}


def standardize_model_name(model_name: str, is_completion: bool = False) -> str:
    """
    Standardize the model name to a format that can be used in the OpenAI API.

    Args:
        model_name: Model name to standardize.
        is_completion: Whether the model is used for completion or not.
            Defaults to False.

    Returns:
        Standardized model name.

    """
    model_name = model_name.lower()
    if ".ft-" in model_name:
        model_name = model_name.split(".ft-")[0] + "-azure-finetuned"
    if ":ft-" in model_name:
        model_name = model_name.split(":")[0] + "-finetuned-legacy"
    if "ft:" in model_name:
        model_name = model_name.split(":")[1] + "-finetuned"
    if model_name == "text-embedding-ada-002-v2":
        model_name = "text-embedding-ada-002"
    if is_completion and (
        model_name.startswith("gpt-4")
        or model_name.startswith("gpt-3.5")
        or model_name.startswith("gpt-35")
        or ("finetuned" in model_name and "legacy" not in model_name)
    ):
        return model_name + "-completion"
    else:
        return model_name


def get_openai_token_cost_for_model(
    model_name: str, num_tokens: int, is_completion: bool = False
) -> float:
    """
    Get the cost in USD for a given model and number of tokens.

    Args:
        model_name: Name of the model
        num_tokens: Number of tokens.
        is_completion: Whether the model is used for completion or not.
            Defaults to False.

    Returns:
        Cost in USD.
    """

    if not model_name:
        logger.warning("failed to get token cost for model name is none")
        return 0

    if not num_tokens:
        logger.warning("failed to get token cost for num_tokens is zero")
        return 0

    model_name = standardize_model_name(model_name, is_completion=is_completion)
    if model_name not in MODEL_COST_PER_1K_TOKENS:
        logger.warning(f"failed to get token cost for '{model_name}' is unsupported")
        return 0
    cost = MODEL_COST_PER_1K_TOKENS[model_name] * (num_tokens / 1000)
    return round(cost, 6)


def num_tokens_from_messages(
    messages: Union[str, List[Any]], model="gpt-3.5-turbo-0613"
) -> int:
    """
    Return the number of tokens used the messages.

    NOTE: this function won't assure exact the same result of OpenAI

    Refer: https://github.com/openai/openai-cookbook/blob/main/examples/How_to_count_tokens_with_tiktoken.ipynb
    """
    import tiktoken

    if not model:
        logger.warning("failed to calculate tokens for message for model name is none")
        return 0

    try:
        encoding = tiktoken.encoding_for_model(model)
    except Exception:
        logger.warning(f"{model} not found. Using cl100k_base encoding.")
        encoding = tiktoken.get_encoding("cl100k_base")
    if isinstance(messages, str):
        return len(encoding.encode(messages))

    if model in {
        "gpt-3.5-turbo-0613",
        "gpt-3.5-turbo-16k-0613",
        "gpt-4-0314",
        "gpt-4-32k-0314",
        "gpt-4-0613",
        "gpt-4-32k-0613",
    }:
        tokens_per_message = 3
        tokens_per_name = 1
    elif model == "gpt-3.5-turbo-0301":
        tokens_per_message = (
            4  # every message follows <|start|>{role/name}\n{content}<|end|>\n
        )
        tokens_per_name = -1  # if there's a name, the role is omitted
    elif "gpt-3.5-turbo" in model:
        logger.warning(
            "gpt-3.5-turbo may update over time. Returning num tokens assuming gpt-3.5-turbo-0613."
        )
        return num_tokens_from_messages(messages, model="gpt-3.5-turbo-0613")
    elif "gpt-4" in model:
        logger.warning(
            "gpt-4 may update over time. Returning num tokens assuming gpt-4-0613."
        )
        return num_tokens_from_messages(messages, model="gpt-4-0613")
    else:
        logger.warning(
            f"greptimeai doesn't support the computation of tokens for {model} at this time."
        )
        return 0
    num_tokens = 0
    for message in messages:
        num_tokens += tokens_per_message
        for key, value in message.items():
            num_tokens += len(encoding.encode(value))
            if key == "name":
                num_tokens += tokens_per_name
    num_tokens += 3  # every reply is primed with <|start|>assistant<|message|>
    return num_tokens


AUDIO_MODEL_COST_PER_1K_CHARS = {
    "tts-1": 0.015,
    "tts-1-hd": 0.03,
}


def num_characters_for_audio(input: str) -> int:
    """
    The maximum length is 4096 characters.

    Refer: https://platform.openai.com/docs/api-reference/audio/createSpeech
    """
    if not input:
        logger.warning("failed to get audio cost for input is none")
        return 0

    return min(len(input), 4096)


def get_openai_audio_cost_for_tts(model_name: str, num_chars: int) -> float:
    """
    NOTE: this function won't assure exact the same result of OpenAI

    Refer: https://platform.openai.com/docs/api-reference/audio/createSpeech
    """
    if model_name not in AUDIO_MODEL_COST_PER_1K_CHARS:
        logger.warning(f"failed to get ttl cost for '{model_name}' is unsupported")
        return 0

    cost = AUDIO_MODEL_COST_PER_1K_CHARS[model_name] * (num_chars / 1000)
    return round(cost, 6)


def get_openai_audio_cost_for_whisper(seconds: int) -> float:
    """
    cost: 0.006 per minute, rounded to the nearest second
    """
    cost = 0.006 * (seconds / 60)
    return round(cost, 6)


def extract_chat_inputs(messages: List[ChatCompletionMessageParam]) -> str:
    """
    this is for display the inputs in the UI.

    NOTE: DO NOT support completion, which will be shut off on January 4th, 2024.
    """

    if not isinstance(messages, list):
        logger.warning(f"failed to extract chat inputs for {messages} is not a list")
        return ""

    if not all(isinstance(message, dict) for message in messages):
        logger.warning(
            f"failed to extract chat inputs for {messages} is not a list of dict"
        )
        return ""

    def extract_input(message: ChatCompletionMessageParam) -> str:
        role = message.get("role", "")
        content = message.get("content", "")
        return f"{role}: {str(content)}"  # content may not be a str

    return "\n".join([extract_input(message) for message in messages])


def extract_chat_outputs(completion: dict) -> str:
    """
    this is for display the outputs in the UI

    NOTE: DO NOT support completion, which will be shut off on January 4th, 2024.
    """

    def extract_choice(choice: dict) -> str:
        message = choice.get("message", {})
        role = message.get("role", "")
        content = message.get("content", "")
        return f"{role}: {str(content)}"  # content may not be a str

    return "\n".join(
        [extract_choice(choice) for choice in completion.get("choices", [])]
    )



from typing import Union

import openai
from openai import AsyncOpenAI, OpenAI

from greptimeai.patchee import Patchee

from . import _SPAN_NAME_FILE, OpenaiPatchees


class _FilePatchees:
    def __init__(self, client: Union[OpenAI, AsyncOpenAI, None], method_name: str):
        files = Patchee(
            obj=client.files if client else openai.files,
            method_name=method_name,
            span_name=_SPAN_NAME_FILE,
            event_name=f"files.{method_name}",
        )

        files_raw = Patchee(
            obj=client.files.with_raw_response
            if client
            else openai.files.with_raw_response,
            method_name=method_name,
            span_name=_SPAN_NAME_FILE,
            event_name=f"files.with_raw_response.{method_name}",
        )

        self.patchees = [files, files_raw]

        if client:
            raw_files = Patchee(
                obj=client.with_raw_response.files,
                method_name=method_name,
                span_name=_SPAN_NAME_FILE,
                event_name=f"with_raw_response.files.{method_name}",
            )
            self.patchees.append(raw_files)


class _FileListPatchees(_FilePatchees):
    def __init__(self, client: Union[OpenAI, AsyncOpenAI, None] = None):
        super().__init__(client=client, method_name="list")


class _FileCreatePatchees(_FilePatchees):
    def __init__(self, client: Union[OpenAI, AsyncOpenAI, None] = None):
        super().__init__(client=client, method_name="create")


class _FileDeletePatchees(_FilePatchees):
    def __init__(self, client: Union[OpenAI, AsyncOpenAI, None] = None):
        super().__init__(client=client, method_name="delete")


class _FileRetrievePatchees(_FilePatchees):
    def __init__(self, client: Union[OpenAI, AsyncOpenAI, None] = None):
        super().__init__(client=client, method_name="retrieve")


class _FileRetrieveContentPatchees(_FilePatchees):
    """
    `.retrieve_content()` is deprecated, but it should be patched as well.
    The `.content()` method should be used instead
    """

    def __init__(self, client: Union[OpenAI, AsyncOpenAI, None] = None):
        super().__init__(client=client, method_name="retrieve_content")


class _FileContentPatchees(_FilePatchees):
    def __init__(self, client: Union[OpenAI, AsyncOpenAI, None] = None):
        super().__init__(client=client, method_name="content")


class FilePatchees(OpenaiPatchees):
    def __init__(self, client: Union[OpenAI, AsyncOpenAI, None] = None):
        list = _FileListPatchees(client=client)
        create = _FileCreatePatchees(client=client)
        delete = _FileDeletePatchees(client=client)
        retrieve = _FileRetrievePatchees(client=client)
        retrieve_content = _FileRetrieveContentPatchees(client=client)
        content = _FileContentPatchees(client=client)

        patchees = (
            list.patchees
            + create.patchees
            + delete.patchees
            + retrieve.patchees
            + retrieve_content.patchees
            + content.patchees
        )

        super().__init__(patchees=patchees, client=client)



from typing import Any, Dict, List, Optional, Sequence, Union
from uuid import UUID

from langchain.callbacks.base import BaseCallbackHandler
from langchain.schema.agent import AgentAction, AgentFinish
from langchain.schema.document import Document
from langchain.schema.messages import BaseMessage, get_buffer_string
from langchain.schema.output import (
    ChatGenerationChunk,
    Generation,
    GenerationChunk,
    LLMResult,
)
from tenacity import RetryCallState

from greptimeai import logger
from greptimeai.collector import Collector
from greptimeai.labels import (
    _CLASS_TYPE_LABEL,
    _COMPLETION_COST_LABEL,
    _COMPLETION_TOKENS_LABEL,
    _ERROR_TYPE_LABEL,
    _INPUT_DISPLAY_LABEL,
    _MODEL_LABEL,
    _OUTPUT_DISPLAY_LABEL,
    _SPAN_NAME_LABEL,
    _USER_ID_LABEL,
    _PROMPT_COST_LABEl,
    _PROMPT_TOKENS_LABEl,
)
from greptimeai.utils.openai.token import (
    get_openai_token_cost_for_model,
    num_tokens_from_messages,
    standardize_model_name,
)

from . import (
    _SPAN_NAME_AGENT,
    _SPAN_NAME_CHAIN,
    _SPAN_NAME_LLM,
    _SPAN_NAME_RETRIEVER,
    _SPAN_NAME_TOOL,
    _get_serialized_id,
    _get_serialized_streaming,
    _get_user_id,
    _parse_documents,
    _parse_generations,
    _parse_input,
    _parse_output,
)


class GreptimeCallbackHandler(Collector, BaseCallbackHandler):
    """
    Greptime LangChain callback handler to collect metrics and traces.
    """

    def __init__(
        self,
        host: str = "",
        database: str = "",
        token: str = "",
        verbose: bool = True,
    ):
        super().__init__(
            service_name="langchain", host=host, database=database, token=token
        )
        self._verbose = verbose

    def on_chain_start(
        self,
        serialized: Dict[str, Any],
        inputs: Dict[str, Any],
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        tags: Union[List[str], None] = None,
        metadata: Union[Dict[str, Any], None] = None,
        **kwargs: Any,
    ) -> Any:
        logger.debug(
            f"on_chain_start. { run_id =} { parent_run_id =} { kwargs = } { serialized = }"
        )

        span_attrs = {
            _USER_ID_LABEL: _get_user_id(metadata),
        }

        event_attrs = {
            _CLASS_TYPE_LABEL: _get_serialized_id(serialized),
            "metadata": metadata,
            "tags": tags,
        }
        if self._verbose:
            event_attrs["inputs"] = _parse_input(inputs)

        self._collector.start_span(
            span_id=run_id,
            parent_id=parent_run_id,
            span_name=_SPAN_NAME_CHAIN,
            event_name="chain_start",
            span_attrs=span_attrs,
            event_attrs=event_attrs,
        )

    def on_chain_end(
        self,
        outputs: Dict[str, Any],
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        **kwargs: Any,
    ) -> Any:
        logger.debug(f"on_chain_end. { run_id =} { parent_run_id =} { kwargs = }")
        event_attrs = {}
        if self._verbose:
            event_attrs["outputs"] = _parse_output(outputs)

        self._collector.end_span(
            span_id=run_id,
            span_name=_SPAN_NAME_CHAIN,
            span_attrs={},
            event_name="chain_end",
            event_attrs=event_attrs,
        )

    def on_chain_error(
        self,
        error: BaseException,
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        **kwargs: Any,
    ) -> Any:
        logger.debug(f"on_chain_error. { run_id =} { parent_run_id =} { kwargs = }")
        event_attrs = {
            _ERROR_TYPE_LABEL: error.__class__.__name__,
        }

        self._collector.end_span(
            span_id=run_id,
            span_name=_SPAN_NAME_CHAIN,
            span_attrs={},
            event_name="chain_error",
            event_attrs=event_attrs,
            ex=error,
        )
        self._collector.collect_error_count(
            {
                _SPAN_NAME_LABEL: _SPAN_NAME_CHAIN,
                **event_attrs,
            },
        )

    def on_llm_start(
        self,
        serialized: Dict[str, Any],
        prompts: List[str],
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        tags: Union[List[str], None] = None,
        metadata: Union[Dict[str, Any], None] = None,
        invocation_params: Union[Dict[str, Any], None] = None,
        **kwargs: Any,
    ) -> Any:
        logger.debug(
            f"on_llm_start. { run_id =} { parent_run_id =} { kwargs = } { serialized = }"
        )

        inputs = " ".join(prompts)
        span_attrs: Dict[str, Any] = {
            _USER_ID_LABEL: _get_user_id(metadata),
            _INPUT_DISPLAY_LABEL: inputs,
        }

        streaming = _get_serialized_streaming(serialized)
        if streaming and invocation_params:
            model_name: str = invocation_params.get("model_name", "")
            prompt_tokens = num_tokens_from_messages(inputs, model_name)
            prompt_cost = get_openai_token_cost_for_model(model_name, prompt_tokens)
            span_attrs[_MODEL_LABEL] = model_name
            span_attrs[_PROMPT_TOKENS_LABEl] = prompt_tokens
            span_attrs[_PROMPT_COST_LABEl] = prompt_cost

        event_attrs = {
            _CLASS_TYPE_LABEL: _get_serialized_id(serialized),
            "metadata": metadata,
            "tags": tags,
            "params": invocation_params,
        }
        if self._verbose:
            event_attrs["prompts"] = prompts

        self._collector.start_latency(run_id, _SPAN_NAME_LLM)
        self._collector.start_span(
            span_id=run_id,
            parent_id=parent_run_id,
            span_name=_SPAN_NAME_LLM,
            event_name="llm_start",
            span_attrs=span_attrs,
            event_attrs=event_attrs,
        )

    def on_chat_model_start(
        self,
        serialized: Dict[str, Any],
        messages: List[List[BaseMessage]],
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        tags: Union[List[str], None] = None,
        metadata: Union[Dict[str, Any], None] = None,
        invocation_params: Union[Dict[str, Any], None] = None,
        **kwargs: Any,
    ) -> Any:
        logger.debug(
            f"on_chat_model_start. { run_id =} { parent_run_id =} { kwargs = } { serialized = }"
        )

        inputs = get_buffer_string(messages[0])
        span_attrs: Dict[str, Any] = {
            _USER_ID_LABEL: _get_user_id(metadata),
        }
        if self._verbose:
            span_attrs[_INPUT_DISPLAY_LABEL] = inputs

        streaming = _get_serialized_streaming(serialized)
        if streaming and invocation_params:
            model_name: str = invocation_params.get("model_name", "")
            prompt_tokens = num_tokens_from_messages(inputs, model_name)
            prompt_cost = get_openai_token_cost_for_model(model_name, prompt_tokens)
            span_attrs[_MODEL_LABEL] = model_name
            span_attrs[_PROMPT_TOKENS_LABEl] = prompt_tokens
            span_attrs[_PROMPT_COST_LABEl] = prompt_cost

        event_attrs = {
            _CLASS_TYPE_LABEL: _get_serialized_id(serialized),
            "metadata": metadata,
            "tags": tags,
            "params": invocation_params,
        }
        if self._verbose:
            event_attrs["messages"] = inputs

        self._collector.start_latency(run_id, _SPAN_NAME_LLM)
        self._collector.start_span(
            span_id=run_id,
            parent_id=parent_run_id,
            span_name=_SPAN_NAME_LLM,
            event_name="chat_model_start",
            span_attrs=span_attrs,
            event_attrs=event_attrs,
        )

    def on_llm_end(
        self,
        response: LLMResult,
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        **kwargs: Any,
    ) -> Any:
        logger.debug(
            f"on_llm_end. { run_id =} { parent_run_id =} { kwargs = } { response = }"
        )
        generations: List[Generation] = (
            response.generations[0]
            if response and len(response.generations) > 0
            else []
        )
        texts = [generation.text for generation in generations]
        outputs = " ".join(texts)

        output = response.llm_output or {}
        model_name: Optional[str] = output.get("model_name")
        if model_name is None:
            model_name = self._collector.get_model_in_context(run_id) or ""
        model_name = standardize_model_name(model_name)

        token_usage = output.get("token_usage", {})

        # NOTE: only cost of OpenAI model will be calculated and collected so far
        prompt_tokens, prompt_cost, completion_tokens, completion_cost = 0, 0.0, 0, 0.0
        if model_name.strip() != "":
            if len(token_usage) > 0:
                prompt_tokens = token_usage.get("prompt_tokens", 0)
                completion_tokens = token_usage.get("completion_tokens", 0)
                prompt_cost = get_openai_token_cost_for_model(model_name, prompt_tokens)
                completion_cost = get_openai_token_cost_for_model(
                    model_name, completion_tokens, is_completion=True
                )
            else:  # streaming
                completion_tokens = num_tokens_from_messages(outputs, model_name)
                completion_cost = get_openai_token_cost_for_model(
                    model_name, completion_tokens, is_completion=True
                )

        attributes = {_MODEL_LABEL: model_name, _SPAN_NAME_LABEL: _SPAN_NAME_LLM}
        self._collector.collect_metrics(
            prompt_tokens=prompt_tokens,
            prompt_cost=prompt_cost,
            completion_tokens=completion_tokens,
            completion_cost=completion_cost,
            attrs=attributes,
        )

        common_attrs: Dict[str, Any] = {}

        if model_name.strip() != "":
            common_attrs[_MODEL_LABEL] = model_name
        if prompt_tokens > 0:
            common_attrs[_PROMPT_TOKENS_LABEl] = prompt_tokens
        if prompt_cost > 0:
            common_attrs[_PROMPT_COST_LABEl] = prompt_cost
        if completion_tokens > 0:
            common_attrs[_COMPLETION_TOKENS_LABEL] = completion_tokens
        if completion_cost > 0:
            common_attrs[_COMPLETION_COST_LABEL] = completion_cost

        span_attrs = {
            _OUTPUT_DISPLAY_LABEL: outputs,
            **common_attrs,
        }

        event_attrs = common_attrs.copy()
        if self._verbose:
            event_attrs["outputs"] = _parse_generations(generations)

        self._collector.end_latency(run_id, _SPAN_NAME_LLM, attributes=attributes)
        self._collector.end_span(
            span_id=run_id,
            span_name=_SPAN_NAME_LLM,
            span_attrs=span_attrs,
            event_name="llm_end",
            event_attrs=event_attrs,
        )

    def on_llm_error(
        self,
        error: BaseException,
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        **kwargs: Any,
    ) -> Any:
        logger.debug(f"on_llm_error. { run_id =} { parent_run_id =} { kwargs = }")
        event_attrs = {
            _ERROR_TYPE_LABEL: error.__class__.__name__,
        }

        self._collector.end_span(
            span_id=run_id,
            span_name=_SPAN_NAME_LLM,
            span_attrs={},
            event_name="llm_error",
            event_attrs=event_attrs,
            ex=error,
        )
        self._collector.discard_latency(run_id, _SPAN_NAME_LLM)
        self._collector.collect_error_count(
            {
                _SPAN_NAME_LABEL: _SPAN_NAME_LLM,
                **event_attrs,
            },
        )

    def on_llm_new_token(
        self,
        token: str,
        *,
        chunk: Optional[Union[GenerationChunk, ChatGenerationChunk]] = None,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        **kwargs: Any,
    ) -> Any:
        """
        on_llm_start, or on_chat_model_start has already started this span.
        """
        logger.debug(
            f"on_llm_new_token. { run_id = } { parent_run_id = } { kwargs = } { token = } { chunk = }"
        )

        event_attrs = {}
        if self._verbose:
            event_attrs["token"] = token

        self._collector.add_span_event(
            span_id=run_id, event_name="streaming", event_attrs=event_attrs
        )

    def on_tool_start(
        self,
        serialized: Dict[str, Any],
        input_str: str,
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        tags: Union[List[str], None] = None,
        metadata: Union[Dict[str, Any], None] = None,
        **kwargs: Any,
    ) -> Any:
        logger.debug(
            f"on_tool_start. { run_id = } { parent_run_id = } { kwargs = } { serialized = }"
        )

        span_attrs = {
            _USER_ID_LABEL: _get_user_id(metadata),
        }
        event_attrs = {
            _CLASS_TYPE_LABEL: _get_serialized_id(serialized),
            "tool_name": serialized.get("name"),
            "tags": tags,
            "metadata": metadata,
        }
        if self._verbose:
            event_attrs["input"] = input_str

        self._collector.start_span(
            span_id=run_id,
            parent_id=parent_run_id,
            span_name=_SPAN_NAME_TOOL,
            event_name="tool_start",
            span_attrs=span_attrs,
            event_attrs=event_attrs,
        )

    def on_tool_end(
        self,
        output: str,
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        **kwargs: Any,
    ) -> Any:
        logger.debug(f"on_tool_end. { run_id =} { parent_run_id =} { kwargs = }")
        event_attrs = {}
        if self._verbose:
            event_attrs["output"] = output

        self._collector.end_span(
            span_id=run_id,
            span_name=_SPAN_NAME_TOOL,
            span_attrs={},
            event_name="tool_end",
            event_attrs=event_attrs,
        )

    def on_tool_error(
        self,
        error: BaseException,
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        **kwargs: Any,
    ) -> Any:
        logger.debug(f"on_tool_error. { run_id =} { parent_run_id =} { kwargs = }")
        event_attrs = {
            _ERROR_TYPE_LABEL: error.__class__.__name__,
        }

        self._collector.end_span(
            span_id=run_id,
            span_name=_SPAN_NAME_TOOL,
            span_attrs={},
            event_name="tool_error",
            event_attrs=event_attrs,
            ex=error,
        )
        self._collector.collect_error_count(
            {
                _SPAN_NAME_LABEL: _SPAN_NAME_TOOL,
                **event_attrs,
            },
        )

    def on_agent_action(
        self,
        action: AgentAction,
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        tags: Union[List[str], None] = None,
        metadata: Union[Dict[str, Any], None] = None,
        **kwargs: Any,
    ) -> Any:
        logger.debug(f"on_agent_action. { run_id =} { parent_run_id =} { kwargs = }")

        span_attrs = {
            _USER_ID_LABEL: _get_user_id(metadata),
        }

        event_attrs = {
            "tool": action.tool,
            "tool_input": action.tool_input,
            _CLASS_TYPE_LABEL: action.__class__.__name__,
            "log": action.log,
            "tags": tags,
            "metadata": metadata,
        }
        if self._verbose:
            event_attrs["input"] = _parse_input(action.tool_input)

        self._collector.start_span(
            span_id=run_id,
            parent_id=parent_run_id,
            span_name=_SPAN_NAME_AGENT,
            event_name="agent_action",
            span_attrs=span_attrs,
            event_attrs=event_attrs,
        )

    def on_agent_finish(
        self,
        finish: AgentFinish,
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        **kwargs: Any,
    ) -> Any:
        logger.debug(f"on_agent_finish. { run_id =} { parent_run_id =} { kwargs = }")
        event_attrs = {
            _CLASS_TYPE_LABEL: finish.__class__.__name__,
            "log": finish.log,
        }
        if self._verbose:
            event_attrs["output"] = _parse_output(finish.return_values)

        self._collector.end_span(
            span_id=run_id,
            span_name=_SPAN_NAME_AGENT,
            span_attrs={},
            event_name="agent_finish",
            event_attrs=event_attrs,
        )

    def on_retriever_start(
        self,
        serialized: Dict[str, Any],
        query: str,
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        tags: Optional[List[str]] = None,
        metadata: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ) -> Any:
        logger.debug(
            f"on_retriever_start. { run_id = } { parent_run_id = } { kwargs = } { serialized = }"
        )

        span_attrs = {
            _USER_ID_LABEL: _get_user_id(metadata),
        }
        event_attrs = {
            _CLASS_TYPE_LABEL: _get_serialized_id(serialized),
            "tags": tags,
            "metadata": metadata,
        }
        if self._verbose:
            event_attrs["query"] = query

        self._collector.start_span(
            span_id=run_id,
            parent_id=parent_run_id,
            span_name=_SPAN_NAME_RETRIEVER,
            event_name="retriever_start",
            span_attrs=span_attrs,
            event_attrs=event_attrs,
        )

    def on_retriever_error(
        self,
        error: BaseException,
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        **kwargs: Any,
    ) -> Any:
        logger.debug(f"on_retriever_error. {run_id=} {parent_run_id=} {kwargs=}")
        event_attrs = {
            _ERROR_TYPE_LABEL: error.__class__.__name__,
        }
        self._collector.end_span(
            span_id=run_id,
            span_name=_SPAN_NAME_RETRIEVER,
            span_attrs={},
            event_name="retriever_error",
            event_attrs=event_attrs,
            ex=error,
        )
        self._collector.collect_error_count(
            {
                _SPAN_NAME_LABEL: _SPAN_NAME_RETRIEVER,
                **event_attrs,
            },
        )

    def on_retriever_end(
        self,
        documents: Sequence[Document],
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        tags: Optional[List[str]] = None,
        **kwargs: Any,
    ) -> Any:
        logger.debug(f"on_retriever_end. {run_id=} {parent_run_id=} {kwargs=}")
        event_attrs: Dict[str, Any] = {
            "tags": tags,
        }
        if self._verbose:
            event_attrs["docs"] = _parse_documents(documents)

        self._collector.end_span(
            span_id=run_id,
            span_name=_SPAN_NAME_RETRIEVER,
            span_attrs={},
            event_name="retriever_end",
            event_attrs=event_attrs,
        )

    def on_retry(
        self,
        retry_state: RetryCallState,
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        **kwargs: Any,
    ) -> Any:
        logger.debug(f"on_retriever_end. {run_id=} {parent_run_id=} {kwargs=}")
        event_attrs = {
            "retry_state": f"{retry_state}",
        }
        self._collector.add_span_event(
            span_id=run_id, event_name="retry", event_attrs=event_attrs
        )


__all__ = ["GreptimeCallbackHandler"]



from typing import Union

import openai
from openai import AsyncOpenAI, OpenAI

from greptimeai.patchee import Patchee

from . import _SPAN_NAME_MODERATION, OpenaiPatchees


class ModerationPatchees(OpenaiPatchees):
    def __init__(self, client: Union[OpenAI, AsyncOpenAI, None] = None):
        moderations_create = Patchee(
            obj=client.moderations if client else openai.moderations,
            method_name="create",
            span_name=_SPAN_NAME_MODERATION,
            event_name="moderations.create",
        )

        moderations_raw_create = Patchee(
            obj=client.moderations.with_raw_response
            if client
            else openai.moderations.with_raw_response,
            method_name="create",
            span_name=_SPAN_NAME_MODERATION,
            event_name="moderations.with_raw_response.create",
        )

        patchees = [moderations_create, moderations_raw_create]

        if client:
            raw_moderations_create = Patchee(
                obj=client.with_raw_response.moderations,
                method_name="create",
                span_name=_SPAN_NAME_MODERATION,
                event_name="with_raw_response.moderations.create",
            )
            patchees.append(raw_moderations_create)

        super().__init__(patchees=patchees, client=client)



import base64
import json
import logging
import os
import time
from typing import Any, Dict, List, Optional, Tuple, Union
from uuid import UUID

from opentelemetry import metrics, trace
from opentelemetry.context.context import Context
from opentelemetry.exporter.otlp.proto.http.metric_exporter import OTLPMetricExporter
from opentelemetry.exporter.otlp.proto.http.trace_exporter import OTLPSpanExporter
from opentelemetry.metrics import CallbackOptions, Observation
from opentelemetry.sdk.metrics import MeterProvider
from opentelemetry.sdk.metrics.export import PeriodicExportingMetricReader
from opentelemetry.sdk.resources import SERVICE_NAME, Resource
from opentelemetry.sdk.trace import TracerProvider
from opentelemetry.sdk.trace.export import BatchSpanProcessor
from opentelemetry.trace import Span, Status, StatusCode, Tracer, set_span_in_context
from opentelemetry.trace.span import format_span_id, format_trace_id
from opentelemetry.util.types import Attributes, AttributeValue
from typing_extensions import override

from greptimeai import _NAME, _VERSION, logger

_JSON_KEYS_IN_OTLP_ATTRIBUTES = "otlp_json_keys"

_GREPTIME_HOST_ENV_NAME = "GREPTIMEAI_HOST"
_GREPTIME_DATABASE_ENV_NAME = "GREPTIMEAI_DATABASE"
_GREPTIME_TOKEN_ENV_NAME = "GREPTIMEAI_TOKEN"


def _prefix_with_scheme_if_not_found(endpoint: Optional[str]) -> Optional[str]:
    if not endpoint:
        return endpoint

    endpoint = endpoint.strip()

    if (
        endpoint == ""
        or endpoint.startswith("https://")
        or endpoint.startswith("http://")
    ):
        return endpoint

    return f"https://{endpoint}"


def _extract_token(token: Optional[str]) -> Tuple[str, str]:
    """
    if token is invalid or empty, then invalid auth header will be included
    """
    if token is None or token.strip() == "":
        return "", ""

    lst = token.split(":", 2)
    if len(lst) != 2:
        return "", ""

    return lst[0], lst[1]


def _check_with_env(
    name: str, value: Optional[str], env_name: str, required: bool = True
) -> str:
    if value and value.strip() != "":
        return value

    value = os.getenv(env_name, "")

    if required and not value:
        raise ValueError(
            f"{name} MUST BE provided either by passing arguments or setup environment variable {env_name}"
        )

    return value


def _is_valid_otel_attributes_value_type(val: Any) -> bool:
    """
    check if value is valid type for OTel attributes
    """
    return isinstance(val, (bool, str, int, float, bytes))


def _sanitate_attributes(attrs: Optional[Dict[str, Any]]) -> Dict[str, AttributeValue]:
    """
    prepare attributes value to any of ['bool', 'str', 'bytes', 'int', 'float']
    or a sequence of these types.

    Other types will be sanitated to json string, and
    append this key in `_JSON_KEYS_IN_OTLP_ATTRIBUTES`
    """
    result: Dict[str, AttributeValue] = {}
    if not attrs:
        return result

    def _sanitate_list(lst: list) -> Union[list, str]:
        contains_any_invalid_value_type = False
        for item in lst:
            if not _is_valid_otel_attributes_value_type(item):
                contains_any_invalid_value_type = True
                break
        if contains_any_invalid_value_type:
            return json.dumps(lst)
        else:
            return lst

    json_keys = []
    for key, val in attrs.items():
        if val is None:
            continue
        elif _is_valid_otel_attributes_value_type(val):
            result[key] = val
        elif isinstance(val, list):
            sanitated_lst = _sanitate_list(val)
            result[key] = sanitated_lst

            if isinstance(sanitated_lst, str):
                json_keys.append(key)
        else:
            result[key] = json.dumps(val)
            json_keys.append(key)

    if len(json_keys) > 0:
        result[_JSON_KEYS_IN_OTLP_ATTRIBUTES] = json_keys
    return result


class _SpanContext:
    """
    ease context management for OTLP Context and span_id
    """

    def __init__(self, name: str, model: str, span: Span):
        self.name = name
        self.model = model
        self.span = span

    def set_self_as_current(self) -> Context:
        """
        call `get_current_span` from the returning context to retrieve the span
        """
        return set_span_in_context(self.span, Context({}))

    @override
    def __repr__(self) -> str:
        span_context = self.span.get_span_context()
        return (
            f"{self.name}.{self.model}.{span_context.trace_id}.{span_context.span_id}"
        )


class _SpanTable:
    """
    NOTE: different span_name may have same span_id.
    """

    def __init__(self):
        self._spans: Dict[str, List[_SpanContext]] = {}

    def put_span_context(self, key: str, context: _SpanContext):
        """
        Pay Attention:

        - different span_name may have same run_id in LangChain.
        - the key may not be the OTLP span_id, maybe it is UUID format from langchain.
          The key is to help find the span_context.
        """
        context_list = self._spans.get(key, [])
        context_list.append(context)
        self._spans[key] = context_list

    def get_span_context(
        self, key: str, span_name: Optional[str] = None
    ) -> Optional[_SpanContext]:
        """
        Args:

            span_name: if is None or empty, the last context will be returned
        """
        context_list = self._spans.get(key, [])
        if len(context_list) == 0:
            return None

        if not span_name:
            return context_list[-1]

        for context in context_list:
            if span_name == context.name:
                return context

        return None

    def pop_span_context(
        self, key: str, span_name: Optional[str] = None
    ) -> Optional[_SpanContext]:
        """
        if there is only one span matched this span_id, then span_id key will be removed

        Args:

            span_name: if is None or empty, the last context will be returned
        """
        context_list = self._spans.get(key, [])

        if len(context_list) == 0:
            return None

        target_context = None
        rest_list = []

        if not span_name:
            target_context = context_list.pop()
            rest_list = context_list
        else:
            for context in context_list:
                if span_name == context.name:
                    target_context = context
                else:
                    rest_list.append(context)

        if len(rest_list) == 0:
            self._spans.pop(key, None)
        else:
            self._spans[key] = rest_list

        return target_context


class _DurationTable:
    def __init__(self):
        self._tables: Dict[str, float] = {}

    def _key(self, span_id: Union[UUID, str], name: Optional[str]) -> str:
        if name:
            return f"{name}.{span_id}"
        else:
            return str(span_id)

    def set(self, span_id: Union[UUID, str], name: Optional[str] = None):
        """
        set start time of span_id.

        Args:

            name: same span may have different name, this is to differentiate them.
        """
        key = self._key(span_id, name)
        self._tables[key] = time.time()

    def latency_in_ms(
        self, span_id: Union[UUID, str], name: Optional[str] = None
    ) -> Optional[float]:
        """
        return latency in millisecond if key exist, None if not exist.
        """
        key = self._key(span_id, name)
        start = self._tables.pop(key, None)
        if start:
            now = time.time()
            return 1000 * (now - start)
        return None


class _Observation:
    """
    # FIXME(yuanbohan): concurrent conflict. refer to Mutex or other solutions
    """

    def __init__(self, name: str):
        self._name = name
        self._value: Dict[Tuple, float] = {}

    def _reset(self):
        self._value = {}

    def _attrs_to_tuple(self, attrs: Optional[Attributes] = None) -> Tuple:
        """
        sort keys first
        """
        if attrs is None or len(attrs) == 0:
            return ()

        lst = [(key, attrs[key]) for key in sorted(attrs.keys())]
        return tuple(lst)

    def put(self, val: float, attrs: Optional[Attributes] = None):
        """
        if attrs is None or empty, nothing will happen
        """
        if attrs is None or len(attrs) == 0:
            logging.info(f"None key for { attrs }")
            return

        tuple_key = self._attrs_to_tuple(attrs)
        self._value.setdefault(tuple_key, 0)
        self._value[tuple_key] += val

    def observation_callback(self):
        """
        _cost will be reset each time being called
        """

        def callback(_: CallbackOptions):
            lst = [
                Observation(val, dict(tuple_key))
                for tuple_key, val in self._value.items()
            ]
            self._reset()
            return lst

        return callback


class _Collector:
    """
    Collector class is responsible for collecting metrics and traces.

    Args:
        service_name (str): The name of the service.
        host (str, optional): The host URL. Defaults to "".
        database (str, optional): The name of the database. Defaults to "".
        token (str, optional): The authentication token. Defaults to "".
    """

    def __init__(
        self,
        service_name: str,
        host: str = "",
        database: str = "",
        token: str = "",
    ):
        self.service_name = service_name
        self.host = _prefix_with_scheme_if_not_found(
            _check_with_env("host", host, _GREPTIME_HOST_ENV_NAME, True)
        )
        self.database = _check_with_env(
            "database", database, _GREPTIME_DATABASE_ENV_NAME, True
        )
        self.token = _check_with_env("token", token, _GREPTIME_TOKEN_ENV_NAME, False)

        self._duration_tables = _DurationTable()
        self._prompt_cost = _Observation("prompt_cost")
        self._completion_cost = _Observation("completion_cost")
        self._span_tables = _SpanTable()

        self._setup_otel_exporter()
        self._setup_otel_metrics()

    def _setup_otel_exporter(self):
        resource = Resource.create({SERVICE_NAME: self.service_name})
        metrics_endpoint = f"{self.host}/v1/otlp/v1/metrics"
        trace_endpoint = f"{self.host}/v1/otlp/v1/traces"

        username, password = _extract_token(self.token)
        auth = f"{username}:{password}"
        b64_auth = base64.b64encode(auth.encode()).decode("ascii")
        greptime_headers: Dict[str, str] = {
            "Authorization": f"Basic {b64_auth}",
            "x-greptime-db-name": self.database,
        }

        metrics_exporter = OTLPMetricExporter(
            endpoint=metrics_endpoint,
            headers=greptime_headers,
            timeout=5,
        )
        self._metric_reader = PeriodicExportingMetricReader(
            metrics_exporter, export_interval_millis=15000
        )
        metre_provider = MeterProvider(
            resource=resource, metric_readers=[self._metric_reader]
        )
        metrics.set_meter_provider(metre_provider)

        trace_provider = TracerProvider(resource=resource)
        self._span_processor = BatchSpanProcessor(
            OTLPSpanExporter(
                endpoint=trace_endpoint,
                headers=greptime_headers,
                timeout=5,
            )
        )
        trace_provider.add_span_processor(self._span_processor)
        trace.set_tracer_provider(trace_provider)

    def _setup_otel_metrics(self):
        """
        setup opentelemetry, and raise Error if something wrong
        """
        self._tracer = trace.get_tracer(
            instrumenting_module_name=_NAME,
            instrumenting_library_version=_VERSION,
        )

        meter = metrics.get_meter(name=_NAME, version=_VERSION)

        self._prompt_tokens_count = meter.create_counter(
            "llm_prompt_tokens",
            description="counts the amount of llm prompt token",
        )

        self._completion_tokens_count = meter.create_counter(
            "llm_completion_tokens",
            description="counts the amount of llm completion token",
        )

        self._llm_error_count = meter.create_counter(
            "llm_errors",
            description="counts the amount of llm errors",
        )

        self._requests_duration_histogram = meter.create_histogram(
            name="llm_request_duration_ms",
            description="duration of requests of llm in milliseconds",
            unit="ms",
        )

        meter.create_observable_gauge(
            callbacks=[self._prompt_cost.observation_callback()],
            name="llm_prompt_tokens_cost",
            description="prompt token cost in US Dollar",
        )

        meter.create_observable_gauge(
            callbacks=[self._completion_cost.observation_callback()],
            name="llm_completion_tokens_cost",
            description="completion token cost in US Dollar",
        )

    def start_span(
        self,
        span_id: Union[UUID, str, None],  # uuid from langchain
        parent_id: Union[UUID, str, None],  # uuid from langchain
        span_name: str,
        event_name: str,
        span_attrs: Dict[str, Any] = {},  # model SHOULD exist in span attrs
        event_attrs: Dict[str, Any] = {},
    ) -> Tuple[str, str]:
        """
        NOTE: end_span MUST BE called with the the same span_id and span_name to revoke the key in trace table.

        if span_id is None when calling start_span, then use the returned span_id for end_span.

        Args:

            span_id: span id. If None, this id will be automatically generated.
            parent_id: parent span id. if None, this is a root span.

        Returns:

            Tuple of trace_id and span_id.

        """
        logger.debug(f"start span for {span_name} with {span_id=} or {parent_id=}")
        span_attributes = _sanitate_attributes(span_attrs)
        event_attributes = _sanitate_attributes(event_attrs)

        def _do_start_span(ctx: Optional[Context] = None) -> Tuple[str, str]:
            span = self._tracer.start_span(
                span_name, context=ctx, attributes=span_attributes
            )
            span.add_event(event_name, attributes=event_attributes)

            trace_context = _SpanContext(
                name=span_name, model=span_attrs.get("model", ""), span=span
            )

            span_context = span.get_span_context()
            str_trace_id, str_span_id = format_trace_id(
                span_context.trace_id
            ), format_span_id(span_context.span_id)

            # NOTE: if span_id is specified, then use it as key
            span_key = str(span_id) if span_id else str_span_id
            self._span_tables.put_span_context(span_key, trace_context)
            return str_trace_id, str_span_id

        if parent_id:
            span_context = self._span_tables.get_span_context(str(parent_id))
            if span_context:
                return _do_start_span(span_context.set_self_as_current())
            else:
                logging.error(
                    f"Parent span of { parent_id } not found, will act as Root span."
                )
                return _do_start_span()

        if span_id is None:
            return _do_start_span()

        # span_context may exist for the same run_id in LangChain. For Example:
        # different Chain triggered in the same trace may have the same run_id.
        span_context = self._span_tables.get_span_context(str(span_id))
        if span_context:
            return _do_start_span(span_context.set_self_as_current())
        else:
            return _do_start_span()

    def add_span_event(
        self, span_id: Union[UUID, str], event_name: str, event_attrs: Dict[str, Any]
    ):
        """
        this is mainly focused on LangChain.
        """
        logger.debug(f"add event for {event_name} with {span_id=}")
        attrs = _sanitate_attributes(event_attrs)
        context = self._span_tables.get_span_context(str(span_id))
        if context:
            context.span.add_event(event_name, attributes=attrs)
        else:
            logging.error(f"{span_id} span not found for {event_name}")

    def end_span(
        self,
        span_id: Union[UUID, str],
        span_name: str,
        event_name: str,
        span_attrs: Dict[str, Any] = {},
        event_attrs: Dict[str, Any] = {},
        ex: Optional[BaseException] = None,
    ):
        logger.debug(f"end span for {span_name} with {span_id=}")

        span_attributes = _sanitate_attributes(span_attrs)
        event_attributes = _sanitate_attributes(event_attrs)

        context = self._span_tables.pop_span_context(str(span_id), span_name)
        if context and context.span:
            span = context.span
            if ex:
                span.record_exception(ex)  # type: ignore
            if span_attributes:
                span.set_attributes(attributes=span_attributes)
            code = StatusCode.ERROR if ex else StatusCode.OK
            span.set_status(Status(code))
            span.add_event(event_name, attributes=event_attributes)
            span.end()
        else:
            logging.error(
                f"unexpected behavior of end_span. context of { span_id } and { span_name } not found."
            )

    def collect_metrics(
        self,
        prompt_tokens: int,
        prompt_cost: float,
        completion_tokens: int,
        completion_cost: float,
        attrs: Optional[Attributes] = None,
    ):
        if prompt_tokens:
            self._prompt_tokens_count.add(prompt_tokens, attrs)

        if prompt_cost:
            self._prompt_cost.put(prompt_cost, attrs)

        if completion_tokens:
            self._completion_tokens_count.add(completion_tokens, attrs)

        if completion_cost:
            self._completion_cost.put(completion_cost, attrs)

    def start_latency(self, span_id: Union[UUID, str], span_name: Optional[str]):
        """
        if latency can not be calculated, call start_latency with span_id and span_name,
        then call end_latency with the same span_id and span_name.

        NOTE: end_latency MUST BE called to revoke the key in duration table.
        """
        self._duration_tables.set(span_id, span_name)

    def end_latency(
        self,
        span_id: Union[UUID, str],
        span_name: Optional[str],
        attributes: Dict[str, Any],
    ):
        latency = self._duration_tables.latency_in_ms(span_id, span_name)
        self.record_latency(latency, attributes)

    def record_latency(
        self, latency: Union[None, int, float], attributes: Optional[Attributes] = None
    ):
        """
        directly record latency in millisecond

        Args:

            latency: millisecond unit
        """
        if latency:
            self._requests_duration_histogram.record(latency, attributes)
        else:
            logger.warning(
                f"latency won't be recorded for None value. attribute is: { attributes }"
            )

    def discard_latency(self, span_id: Union[UUID, str], span_name: Optional[str]):
        self._duration_tables.latency_in_ms(span_id, span_name)

    def collect_error_count(self, attributes: Dict[str, Any]):
        self._llm_error_count.add(1, attributes)

    def get_model_in_context(self, span_id: Union[UUID, str]) -> Optional[str]:
        context = self._span_tables.get_span_context(str(span_id))

        if context:
            return context.model

        return None

    @property
    def tracer(self) -> Tracer:
        return self._tracer

    def _force_flush(self):
        """
        DO NOT call this method to flush metrics and traces, this is only for test cases.
        """
        self._metric_reader.force_flush()
        self._span_processor.force_flush()



from typing import Union

import openai
from openai import AsyncOpenAI, OpenAI

from greptimeai.patchee import Patchee

from . import _SPAN_NAME_EMBEDDING, OpenaiPatchees


class EmbeddingPatchees(OpenaiPatchees):
    def __init__(self, client: Union[OpenAI, AsyncOpenAI, None] = None):
        embeddings_create = Patchee(
            obj=client.embeddings if client else openai.embeddings,
            method_name="create",
            span_name=_SPAN_NAME_EMBEDDING,
            event_name="embeddings.create",
        )

        embeddings_raw_create = Patchee(
            obj=client.embeddings.with_raw_response
            if client
            else openai.embeddings.with_raw_response,
            method_name="create",
            span_name=_SPAN_NAME_EMBEDDING,
            event_name="embeddings.with_raw_response.create",
        )

        patchees = [
            embeddings_create,
            embeddings_raw_create,
        ]

        if client:
            raw_embeddings_create = Patchee(
                obj=client.with_raw_response.embeddings,
                method_name="create",
                span_name=_SPAN_NAME_EMBEDDING,
                event_name="with_raw_response.embeddings.create",
            )
            patchees.append(raw_embeddings_create)

        super().__init__(patchees=patchees, client=client)



from typing import Any, Callable, Optional

from greptimeai import logger

_GREPTIMEAI_WRAPPED = "__GREPTIMEAI_WRAPPED__"


class Patchee:
    def __init__(self, obj: Any, method_name: str, span_name: str, event_name: str):
        self.obj = obj
        self.method_name = method_name
        self.span_name = span_name
        self.event_name = event_name

    def __repr__(self):
        return f"<{self.span_name}> with event name <{self.event_name}>"

    def get_func_name(self) -> str:
        return self.method_name

    def _get_func(self) -> Optional[Callable]:
        return getattr(self.obj, self.method_name)

    def get_unwrapped_func(self) -> Optional[Callable]:
        func = self._get_func()
        if not func:
            logger.warning(f"'{self.obj}' has no function '{self.get_func_name()}'.")
            return None

        if hasattr(func, _GREPTIMEAI_WRAPPED):
            logger.warning(
                f"the function '{self.get_func_name()}' has already been patched."
            )
            return None
        return func

    def wrap_func(self, func: Callable):
        setattr(func, _GREPTIMEAI_WRAPPED, True)
        setattr(self.obj, self.method_name, func)



from typing import Union

import openai
from openai import AsyncOpenAI, OpenAI

from greptimeai.patchee import Patchee

from . import OpenaiPatchees


class RetryPatchees(OpenaiPatchees):
    def __init__(self, client: Union[OpenAI, AsyncOpenAI, None] = None):
        retry = Patchee(
            obj=client or openai._client,
            method_name="_retry_request",
            span_name="",  # retry is only event, so span_name won't be used
            event_name="retry_request",
        )

        super().__init__(patchees=[retry], client=client)
